/*
 * StichingImage.cpp
 *
 *  Created on: Dec 10, 2013
 *      Author: truongnt
 */

#include "StitchingImage.h"

#include "opencv2/calib3d/calib3d.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/features2d/features2d.hpp"
#include "opencv2/nonfree/nonfree.hpp"


#include "hawaii/common/timer.h"
#include "hawaii/common/tracker.h"

#include "base/basic_function.h"

using namespace cv;
using namespace std;

namespace producer_consumer_thread {

StitchingImage::StitchingImage() {
	// TODO Auto-generated constructor stub

}

StitchingImage::~StitchingImage() {
	// TODO Auto-generated destructor stub
}

cv::Point2f StitchingImage::drawOrientationVector(cv::Mat &result, const std::vector<KeyPoint> &keypoints1, const std::vector<KeyPoint> &keypoints2,
		const std::vector< DMatch > &matches,
		const std::vector<char> &matchesMask) {

	Point2f p1Average(0, 0), p2Average(0, 0);
	int iCounter = 0;

    for( size_t m = 0; m < matches.size(); m++ )
    {
        if( matchesMask.empty() || matchesMask[m] )
        {
            int i1 = matches[m].queryIdx;
            int i2 = matches[m].trainIdx;
            CV_Assert(i1 >= 0 && i1 < static_cast<int>(keypoints1.size()));
            CV_Assert(i2 >= 0 && i2 < static_cast<int>(keypoints2.size()));

            const KeyPoint &kp1 = keypoints1[i1], &kp2 = keypoints2[i2];
            //cv::line( result, kp1.pt, kp2.pt, Scalar( 0, 255, 0 ), 2, 8);
            p1Average.x += kp1.pt.x; p1Average.y += kp1.pt.y;
            p2Average.x += kp2.pt.x; p2Average.y += kp2.pt.y;
            iCounter++;
        }
    }

    if (iCounter != 0) {
    	p1Average.x = p1Average.x / (float)iCounter;
    	p1Average.y = p1Average.y / (float)iCounter;
    	p2Average.x = p2Average.x / (float)iCounter;
    	p2Average.y = p2Average.y / (float)iCounter;

    	Point2f pCenterImage(result.size().width / 2.0f, result.size().height / 2.0f);
    	cv::line( result, pCenterImage, pCenterImage + p2Average - p1Average, Scalar( 0, 0, 255 ), 2, 8);
    	return p2Average - p1Average;
    }

    return Point2f(0, 0);
}

//test draw matched by using almost the same technique to draw image, but this time we only draw in one image (look similar to the one of sparse3d)
cv::Mat StitchingImage::drawMatchedPoints(cv::Point2f &orientationVector, Mat img1, Mat img2,
		std::vector<KeyPoint> keypoints1, std::vector<KeyPoint> keypoints2,
		std::vector< DMatch > matches,
		std::vector<char> matchesMask) {
	Mat result = img1.clone();

    // draw matches
    for( size_t m = 0; m < matches.size(); m++ )
    {
        if( matchesMask.empty() || matchesMask[m] )
        {
            int i1 = matches[m].queryIdx;
            int i2 = matches[m].trainIdx;
            CV_Assert(i1 >= 0 && i1 < static_cast<int>(keypoints1.size()));
            CV_Assert(i2 >= 0 && i2 < static_cast<int>(keypoints2.size()));

            const KeyPoint &kp1 = keypoints1[i1], &kp2 = keypoints2[i2];
            cv::line( result, kp1.pt, kp2.pt, Scalar( 0, 255, 0 ), 2, 8);

            //_drawMatch( outImg, outImg1, outImg2, kp1, kp2, matchColor, flags );
        }
    }

    orientationVector = drawOrientationVector(result, keypoints1, keypoints2, matches, matchesMask);

	return result;
}

Mat StitchingImage::getHomographyMatrix3x3(Mat img_1, Mat img_2, Mat &imgAfterRANSAC, Point2f &orientationVector)
{
	try {
		Mat H12;
		string strDetector, strDescriptor, strMatcher;
		strDetector = "SURF";//"ORB";
		strDescriptor = "SURF";//"ORB";
		strMatcher = "";
		//-- Step 1: Detect the keypoints using SURF Detector
		std::vector<KeyPoint> keypoints_1, keypoints_2;
		//-- Step 2: Calculate descriptors (feature vectors)
		Mat descriptors_1, descriptors_2;
		DetectorExtractorMatcher::detectExtractDescriptor(strDetector, strDescriptor,
				img_1, img_2, keypoints_1, keypoints_2, descriptors_1, descriptors_2);
//		DetectKeyPointsSUFT(img_1, img_2, keypoints_1, keypoints_2, descriptors_1, descriptors_2);
//		DetectKeyPoints("ORB", "ORB",
//				img_1, img_2, keypoints_1, keypoints_2, descriptors_1, descriptors_2);

		std::vector< DMatch > matches;
		DetectorExtractorMatcher::matchFeature(strDetector, strDescriptor, strMatcher, keypoints_1, keypoints_2, descriptors_1, descriptors_2, matches);

		////-- Draw  matches
		//Mat img_matches;
		//drawMatches( img_1, keypoints_1, img_2, keypoints_2,
		//	matches, img_matches, Scalar::all(-1), Scalar::all(-1),
		//	vector<char>(), DrawMatchesFlags::NOT_DRAW_SINGLE_POINTS );

		////-- Show detected matches
		//imshow( "After using Matching", img_matches );

		//-----------------------------------------------------
		//using RANSAC in function find homography
		//vector<KeyPoint> keypoints_2;
		//vector<DMatch> filteredMatches;
		vector<int> queryIdxs( matches.size() ), trainIdxs( matches.size() );


		bool isWarpPerspective = false;
		double ransacReprojThreshold = 3;//find what is good

		for( size_t i = 0; i < matches.size(); i++ )
		{
			queryIdxs[i] = matches[i].queryIdx;
			trainIdxs[i] = matches[i].trainIdx;
		}

		if( !isWarpPerspective && ransacReprojThreshold >= 0 )
		{
			//cout << "< Computing homography (RANSAC)..." << endl;
			vector<Point2f> points1; KeyPoint::convert(keypoints_1, points1, queryIdxs);//??
			vector<Point2f> points2; KeyPoint::convert(keypoints_2, points2, trainIdxs);

			H12 = findHomography( Mat(points1), Mat(points2), CV_RANSAC, ransacReprojThreshold );
			//cout << ">" << endl;
		}

		Mat drawImg;
		if( !H12.empty() ) // filter outliers
		{
			std::cout<<"H12.notEmpty()"<<std::endl;
			vector<char> matchesMask( matches.size(), 0 );
			vector<Point2f> points1; KeyPoint::convert(keypoints_1, points1, queryIdxs);
			vector<Point2f> points2; KeyPoint::convert(keypoints_2, points2, trainIdxs);
			Mat points1t; perspectiveTransform(Mat(points1), points1t, H12);

			double maxInlierDist = ransacReprojThreshold < 0 ? 3 : ransacReprojThreshold;
			for( size_t i1 = 0; i1 < points1.size(); i1++ )
			{
				if( norm(points2[i1] - points1t.at<Point2f>((int)i1,0)) <= maxInlierDist ) // inlier
					matchesMask[i1] = 1;
			}
			// draw inliers
//			drawMatches( img_1, keypoints_1, img_2, keypoints_2,
//					matches, drawImg, Scalar::all(-1), Scalar::all(-1),
//					matchesMask);

			drawImg = drawMatchedPoints(orientationVector, img_1, img_2, keypoints_1, keypoints_2, matches, matchesMask);
		}
		else {
			std::cout<<"H12.empty()"<<std::endl;
//			drawMatches( img_1, keypoints_1, img_2, keypoints_2, matches, drawImg );
			drawImg = drawMatchedPoints(orientationVector, img_1, img_2, keypoints_1, keypoints_2, matches);
		}

		cv::drawKeypoints(img_1, keypoints_1, DetailedImg1, Scalar::all(-1), cv::DrawMatchesFlags::DEFAULT);
		cv::drawKeypoints(img_2, keypoints_2, DetailedImg2, Scalar::all(-1), cv::DrawMatchesFlags::DEFAULT);

//		imshow( "After using RANSAC", drawImg );
//		cv::waitKey(1);
		imgAfterRANSAC = drawImg.clone();
		return H12;
	}
	catch (std::exception& e)
	{
		std::cerr << "StichingImage::getHomographyMatrix3x3 " << e.what() << std::endl;
	}

}

Point2f convertNewPotision(Mat RotateMatrix, Point oldPoint)
{
	Point3f newPoint;
	newPoint.x = RotateMatrix.at<double>(0, 0) * (double)oldPoint.x + RotateMatrix.at<double>(0, 1) * (double)oldPoint.y + RotateMatrix.at<double>(0, 2);
	newPoint.y = RotateMatrix.at<double>(1, 0) * (double)oldPoint.x + RotateMatrix.at<double>(1, 1) * (double)oldPoint.y + RotateMatrix.at<double>(1, 2);
	newPoint.z = RotateMatrix.at<double>(2, 0) * (double)oldPoint.x + RotateMatrix.at<double>(2, 1) * (double)oldPoint.y + RotateMatrix.at<double>(2, 2);
	return Point2f(newPoint.x / newPoint.z, newPoint.y / newPoint.z);
}

Vec3b BilinearInterpolation2Point(Vec3b p1, Vec3b p2, float w1, float w2)
{
	return Vec3b(p1[0] * w1 + p2[0] * w2, p1[1] * w1 + p2[1] * w2, p1[2] * w1 + p2[2] * w2);
}

///p1  p4
///
///p2  p3
Vec3b BilinearInterpolation4Point(Vec3b p1, Vec3b p2, Vec3b p3, Vec3b p4, Point2f p_real)
{
	Point2f p_real_weight(p_real.x - (float)((int)p_real.x), p_real.y - (float)((int)p_real.y));

	Vec3b p21 = BilinearInterpolation2Point(p1, p2, p_real_weight.y, 1.0f - p_real_weight.y);
	Vec3b p22 = BilinearInterpolation2Point(p4, p3, p_real_weight.y, 1.0f - p_real_weight.y);
	return BilinearInterpolation2Point(p21, p22, p_real_weight.x, 1.0f - p_real_weight.x);
}

//for grayscale image
double BilinearInterpolation2Point(float p1, float p2, float w1, float w2)
{
	return double(p1 * w1 + p2 * w2);
}

///p1  p4
///
///p2  p3
double BilinearInterpolation4Point(float p1, float p2, float p3, float p4, Point2f p_real)
{
	Point2f p_real_weight(p_real.x - (float)((int)p_real.x), p_real.y - (float)((int)p_real.y));

	float p21 = BilinearInterpolation2Point(p1, p2, p_real_weight.y, 1.0f - p_real_weight.y);
	float p22 = BilinearInterpolation2Point(p4, p3, p_real_weight.y, 1.0f - p_real_weight.y);
	return BilinearInterpolation2Point(p21, p22, p_real_weight.x, 1.0f - p_real_weight.x);
}

Rect findRectangleOfPolygon(vector<Point> pts)
{

	int topHeight, bottomHeight;
	int leftWidth, rightWidth;

	topHeight = pts[0].y;
	bottomHeight = pts[0].y;
	leftWidth = pts[0].x;
	rightWidth = pts[0].x;

	for (unsigned int i = 0; i < pts.size(); i++)
	{
		if (topHeight > pts[i].y) topHeight = pts[i].y;
		if (bottomHeight < pts[i].y) bottomHeight = pts[i].y;
		if (leftWidth > pts[i].x) leftWidth = pts[i].x;
		if (rightWidth < pts[i].x) rightWidth = pts[i].x;
	}

	return Rect(leftWidth, topHeight, (int)(rightWidth - leftWidth), (int)(bottomHeight - topHeight));
}

void PerspectiveMat::warpPerspective2(Mat src, Mat &dst, Mat H12, Point RectConner)
{
	try {
	Mat H21 = H12.inv();
	///do warpperspective by myseft
	for (int y = 0; y < dst.rows; y++)
	{
		for (int x = 0; x < dst.cols; x++)
		{
			//Vec3b Clr = src.at<Vec3b>(r, c);
			///backward
			Point2f oldPos = convertNewPotision(H21, Point(x + RectConner.x, y + RectConner.y));
			if(0 <= oldPos.x && oldPos.x < src.cols
					&& 0 <= oldPos.y && oldPos.y < src.rows)
			{
				//bilinear interpolation
				if(0 < oldPos.x && oldPos.x < src.cols - 1
						&& 0 < oldPos.y && oldPos.y < src.rows - 1)
				{//case 4 point
					///p1  p4
					///  p_real
					///p2  p3
					dst.at<Vec3b>(y, x) = BilinearInterpolation4Point(src.at<Vec3b>((int)(oldPos.y), (int)(oldPos.x)),
							src.at<Vec3b>((int)(oldPos.y + 0.5f), (int)(oldPos.x)),
							src.at<Vec3b>((int)(oldPos.y + 0.5f), (int)(oldPos.x + 0.5f)),
							src.at<Vec3b>((int)(oldPos.y), (int)(oldPos.x + 0.5f)),
							oldPos);
//
//					dst.at<unsigned char>(y, x) =
//							BilinearInterpolation4Point((float)src.at<unsigned char>((int)(oldPos.y), (int)(oldPos.x)),
//									(float)src.at<unsigned char>((int)(oldPos.y + 0.5f), (int)(oldPos.x)),
//									(float)src.at<unsigned char>((int)(oldPos.y + 0.5f), (int)(oldPos.x + 0.5f)),
//									(float)src.at<unsigned char>((int)(oldPos.y), (int)(oldPos.x + 0.5f)),
//							oldPos);
				}
				else
				{
					dst.at<Vec3b>(y, x) = src.at<Vec3b>((int)oldPos.y, (int)oldPos.x);
					//dst.at<Vec3b>(y, x) = Vec3b(0, 0, 255);
				}
			}

		}
	}
	}
	catch (std::exception& e)
	{
		std::cerr << "warpPerspective2 " << e.what() << std::endl;
	}
}
bool PerspectiveMat::create(Mat Src, Mat H12)
{
	try {
		src = Src;
		vector<Point> pts;
		pts.push_back(convertNewPotision(H12, Point(0, 0)));
		pts.push_back(convertNewPotision(H12, Point(src.cols, 0)));
		pts.push_back(convertNewPotision(H12, Point(src.cols, src.rows)));
		pts.push_back(convertNewPotision(H12, Point(0, src.rows)));

		RectConner = findRectangleOfPolygon(pts);
		if (RectConner.width >= 1000 || RectConner.height >= 1000) return false;
		topleftRelative = Point2f(RectConner.x, RectConner.y);
		dst = Mat::zeros(RectConner.height, RectConner.width, src.type());

		vector<Point> pts_with_relative_pos;
		pts_with_relative_pos.push_back(convertNewPotision(H12, Point(0, 0)) - topleftRelative);
		pts_with_relative_pos.push_back(convertNewPotision(H12, Point(src.cols, 0)) - topleftRelative);
		pts_with_relative_pos.push_back(convertNewPotision(H12, Point(src.cols, src.rows)) - topleftRelative);
		pts_with_relative_pos.push_back(convertNewPotision(H12, Point(0, src.rows)) - topleftRelative);

		///dst
		warpPerspective2(src, dst, H12, topleftRelative);

		///dst_msk: find warped-image msk
		dst_msk = cv::Mat::zeros(dst.rows, dst.cols, CV_32FC1);
		fillConvexPoly(dst_msk, pts_with_relative_pos, Scalar::all(1), 8, 0);

		return true;
	}
	catch (std::exception& e)
	{
		std::cerr << "PerspectiveMat::create " << e.what() << std::endl;
	}
}

void PerspectiveMat::drawToFinalImage(Mat H12, Mat &finalImg, Mat &final_msk)
{
	for (int y = 0; y < dst.rows; y++)
	{
		for (int x = 0; x < dst.cols; x++)
		{
			//Vec3b Clr = src.at<Vec3b>(r, c);
			if(dst_msk.at<float>(y, x) > 0.5f)
			{
				if (!(0 <= y && y < finalImg.rows
						&& 0 <= x && x < finalImg.cols)) {
					continue;
				}

				if (final_msk.at<float>(y, x) < 0.5f)
				{
					finalImg.at<Vec3b>(y, x) = dst.at<Vec3b>(y, x);

					final_msk.at<float>(y, x) = final_msk.at<float>(y, x) + 1;
				}
				else
				{//had value in final, find out average color
					Vec3b Clr_final = finalImg.at<Vec3b>(y, x);
					Vec3b Clr_dst = dst.at<Vec3b>(y, x);
					if (Clr_dst == Vec3b(0, 0, 0))
					{
						//final.at<Vec3b>(y, x) = Vec3b(0, 255, 0);
						continue;//TH msk and dst ko sync
					}
					Vec3b Clr;
					float final_msk_relatve = final_msk.at<float>(y, x);

					Clr[0] = ((float)Clr_final[0] * final_msk_relatve + (float)Clr_dst[0]) / (final_msk_relatve + 1);
					Clr[1] = ((float)Clr_final[1] * final_msk_relatve + (float)Clr_dst[1]) / (final_msk_relatve + 1);
					Clr[2] = ((float)Clr_final[2] * final_msk_relatve + (float)Clr_dst[2]) / (final_msk_relatve + 1);

					finalImg.at<Vec3b>(y, x) = Clr;
					final_msk_relatve++;
					final_msk.at<float>(y, x) = final_msk_relatve;
				}
			}
		}
	}

	line(finalImg,
			convertNewPotision(H12, Point(0, 0)) - topleftRelative,
			convertNewPotision(H12, Point(src.cols, 0)) - topleftRelative,
			Scalar(255, 255, 255), 2);

	line(finalImg,
			convertNewPotision(H12, Point(src.cols, 0)) - topleftRelative,
			convertNewPotision(H12, Point(src.cols, src.rows)) - topleftRelative,
			Scalar(255, 255, 255), 2);

	line(finalImg,
			convertNewPotision(H12, Point(src.cols, src.rows)) - topleftRelative,
			convertNewPotision(H12, Point(0, src.rows)) - topleftRelative,
			Scalar(255, 255, 255), 2);

	line(finalImg,
			convertNewPotision(H12, Point(0, src.rows)) - topleftRelative,
			convertNewPotision(H12, Point(0, 0)) - topleftRelative,
			Scalar(255, 255, 255), 2);

}

void StitchingImage::ApplyImage(Mat img_1, Mat img_2)
{
	try {
		if (img_1.empty()) {
			cout<<"Image 1 invalid"<<endl;
			return;
		}

		if (img_2.empty()) {
			cout<<"Image 2 invalid"<<endl;
			return;
		}

		//Mat dst, dst_msk;
		PerspectiveMat dst;
		Mat finalImg;
		Mat final_msk;
		//final = Mat::zeros(img_1.rows * 1.5f, img_1.cols * 1.5f, img_1.type());
		//final_msk = Mat::zeros(img_1.rows * 1.5f, img_1.cols * 1.5f, CV_32FC1);

		Mat H12 = getHomographyMatrix3x3(img_1, img_2, imgAfterRANSAC, OrientationVector);

		//if 1, draw fianl Image, 2 image at the same coordinate
#if 0
		if (!H12.empty()) {
			///at this time, we got H matrix that include rotate and scale element
			///x' = ax + by +c1
			///y' = cx + dy + e1
			//find 4 corner point of image and recaculate SIZE of image
			if (!dst.create(img_1, H12)) return;// the size is too big

			///* warp the image */
			//Mat dst2;
			////warpPerspective(img_1, dst2, H12, Size(img_1.cols * 1.5f, img_1.rows * 1.5f), INTER_LINEAR);
			//warpPerspective(img_1, dst2, H12, Size(img_1.cols, img_1.rows), INTER_LINEAR);
			//imshow("dst2", dst2);

			PerspectiveMat dst2;
			if (!dst2.create(img_2, Mat::eye(Size(3, 3), CV_64FC1))) return;

			///need to find out the best size for final Mat
			vector<Point> pts;
			pts.push_back(dst.topleftRelative);
			pts.push_back(Point(dst.topleftRelative.x + dst.RectConner.width, dst.topleftRelative.y));
			pts.push_back(Point(dst.topleftRelative.x, dst.topleftRelative.y + dst.RectConner.height));
			pts.push_back(Point(dst.topleftRelative.x + dst.RectConner.width, dst.topleftRelative.y + dst.RectConner.height));

			pts.push_back(dst2.topleftRelative);
			pts.push_back(Point(dst2.topleftRelative.x + dst2.RectConner.width, dst2.topleftRelative.y));
			pts.push_back(Point(dst2.topleftRelative.x, dst2.topleftRelative.y + dst2.RectConner.height));
			pts.push_back(Point(dst2.topleftRelative.x + dst2.RectConner.width, dst2.topleftRelative.y + dst2.RectConner.height));

			Rect RectConnerFinal = findRectangleOfPolygon(pts);

			//check size of final image, if it is too big, no need to allocate memory or draw this image!
			if (RectConnerFinal.width >= 1000 || RectConnerFinal.height >= 1000) return;

			finalImg = Mat::zeros(RectConnerFinal.height, RectConnerFinal.width, img_1.type());
			final_msk = Mat::zeros(RectConnerFinal.height, RectConnerFinal.width, CV_32FC1);

			Mat final_dst(finalImg, Rect(dst.RectConner.x - RectConnerFinal.x, dst.RectConner.y - RectConnerFinal.y,
					dst.RectConner.width, dst.RectConner.height));
			Mat final_dst_msk(final_msk, Rect(dst.RectConner.x - RectConnerFinal.x, dst.RectConner.y - RectConnerFinal.y,
					dst.RectConner.width, dst.RectConner.height));

			Mat final_dst2(finalImg, Rect(dst2.RectConner.x - RectConnerFinal.x, dst2.RectConner.y - RectConnerFinal.y,
					dst2.RectConner.width, dst2.RectConner.height));
			Mat final_dst2_msk(final_msk, Rect(dst2.RectConner.x - RectConnerFinal.x, dst2.RectConner.y - RectConnerFinal.y,
					dst2.RectConner.width, dst2.RectConner.height));


			//dst2.drawToFinalImage(final, final_msk);
			//dst.drawToFinalImage(final, final_msk);

			dst2.drawToFinalImage(Mat::eye(Size(3, 3), CV_64FC1), final_dst2, final_dst2_msk);

			//dst.topleftRelative = Point2f(0, 0);
			//dst.RectConner.x = 0;
			//dst.RectConner.y = 0;
			dst.drawToFinalImage(H12, final_dst, final_dst_msk);
			imshow("final2", finalImg);

			//		{
			//			boost::posix_time::ptime current_time = boost::posix_time::microsec_clock::local_time();
			//			boost::posix_time::time_duration diff = current_time - start_time;
			//
			//			std::ostringstream ss;
			//			ss << "data2/stiching_image/stiching_image_"
			//					<< diff.total_milliseconds() << ".jpg";
			//
			//			cv::imwrite(ss.str(), finalImg);
			//		}
			imgFinalResult = finalImg.clone();
		}
#endif
	}
	catch (std::exception& e)
	{
		std::cerr << "StichingImage::ApplyImage "
				<< "line: "<< __LINE__ << " "
				<< e.what() << std::endl;
	}
}

} /* namespace producer_thread */

