/*
 * appLocali.cpp
 *
 *  Created on: 26 nov. 2013
 *      Author: nguyen
 */

#include "appLocali.h"
#include "hawaii/common/optimization.h"
#include "hawaii/common/helpers.h"
#include <opencv2/nonfree/features2d.hpp>
#include <opencv2/nonfree/nonfree.hpp>
#include <opencv2/features2d/features2d.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/calib3d/calib3d.hpp>

#include <stdio.h>
#include <stdlib.h>
#include <string>


DroneAppLocali::DroneAppLocali():
	DroneAppBase(false){

}

DroneAppLocali::~DroneAppLocali() {
	// TODO Auto-generated destructor stub
}

void DroneAppLocali::processImageFront( const cv_bridge::CvImage imageFront ) {
	cv::Mat img = imageFront.image ;
	cv::Mat cannyRes, res;

	cv::Canny(img, cannyRes, 50, 200, 3);
	cv::cvtColor(cannyRes, res, CV_GRAY2BGR);


	cv::imshow("bouh", cannyRes);

	std::vector<cv::Vec4i> lines;
	cv::HoughLinesP(cannyRes, lines, 1, CV_PI/180, 50, 50, 10 );

	for( std::size_t i = 0; i < lines.size(); i++ )
	{
	  cv::Vec4i l = lines[i];
	  cv::line( res, cv::Point(l[0], l[1]), cv::Point(l[2], l[3]), cv::Scalar(0,0,255), 3, CV_AA);
	}

	cv::imshow("detected lines", res);

	cv::waitKey(1);
}

void DroneAppLocali::offlineProc() {
	cv::Mat img1,img2;
	std::string path1("data/bach/hand_light/image313.jpg");
	std::string path2("data/bach/hand_light/image377.jpg");
	img1 = cv::imread( path1, CV_LOAD_IMAGE_GRAYSCALE );
	img2 = cv::imread( path2, CV_LOAD_IMAGE_GRAYSCALE );

	if(!img1.data || !img2.data){
		printf("no image data\n");
		return;
	}

	int minHessian(400);
	cv::SurfFeatureDetector detector(minHessian);
	std::vector<cv::KeyPoint> keypoints1;
	std::vector<cv::KeyPoint> keypoints2;

	detector.detect( img1, keypoints1 );
	detector.detect( img2, keypoints2 );

	cv::Mat img_keypoints_1,img_keypoints_2;
	drawKeypoints( img1, keypoints1, img_keypoints_1, cv::Scalar::all(-1), cv::DrawMatchesFlags::DEFAULT );
	drawKeypoints( img2, keypoints2, img_keypoints_2, cv::Scalar::all(-1), cv::DrawMatchesFlags::DEFAULT );

	cv::imshow("Keypoints 1", img_keypoints_1 );
	cv::imshow("Keypoints 2", img_keypoints_2 );
	cv::imwrite("data/bach/hand_light/out/keypoints313.jpg", img_keypoints_1);
	cv::imwrite("data/bach/hand_light/out/keypoints377.jpg", img_keypoints_2);


	cv::SurfDescriptorExtractor extractor;

	cv::Mat descriptors_1, descriptors_2;

	extractor.compute( img1, keypoints1, descriptors_1 );
	extractor.compute( img2, keypoints2, descriptors_2 );

	cv::FlannBasedMatcher matcher;
	std::vector< cv::DMatch > matches;
	matcher.match( descriptors_1, descriptors_2, matches );

	double max_dist = 0; double min_dist = 100;
	for( int i = 0; i < descriptors_1.rows; i++ )
	{ double dist = matches[i].distance;
		if( dist < min_dist ) min_dist = dist;
		if( dist > max_dist ) max_dist = dist;
	}

	printf("-- Max dist : %f \n", max_dist );
	printf("-- Min dist : %f \n", min_dist );

	double user=getchar()-48;
	printf("-- User : %f \n", user );
	printf("-- Thresh : %f \n", user*min_dist );
	getchar();

	std::vector< cv::DMatch > good_matches;

	for( int i = 0; i < descriptors_1.rows; i++ )
	{ if( matches[i].distance <= user*min_dist )
		{ good_matches.push_back( matches[i]); }
	}


	cv::Mat img_matches;
	cv::drawMatches( img1, keypoints1, img2, keypoints2,
				 good_matches, img_matches, cv::Scalar::all(-1), cv::Scalar::all(-1),
	std::vector<char>(), cv::DrawMatchesFlags::NOT_DRAW_SINGLE_POINTS );

	cv::imshow( "Good Matches", img_matches );
	cv::imwrite("data/bach/hand_light/out/out.jpg",img_matches);

	cv::waitKey(0);
	return;
}
