/*
 * drone_producer.cpp
 *
 *  Created on: Aug 20, 2013
 *      Author: truongnt
 */

#include "drone_producer.h"
#include "hawaii/common/helpers.h"
#include <opencv2/imgproc/imgproc.hpp>
#include <SDL/SDL_keysym.h>
#include <SDL/SDL_joystick.h>

/*
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/fstream.hpp>
#include <fstream>
#include <string>
#include <sstream>
#include <iostream>
 */
#include <unistd.h>
#include <dirent.h>
#include <sys/types.h> // for opendir(), readdir(), closedir()
#include <sys/stat.h> // for stat()

namespace producer_consumer_thread {

AutoNavigator::AutoNavigator(): altitudeDelta(0)
		{
	// TODO Auto-generated constructor stub
	this->state_ = AutoNavigatorState::stateInactive;

}

AutoNavigator::~AutoNavigator() {
	// TODO Auto-generated destructor stub
}

void AutoNavigator::init(const cv::Vec3d      translationGlobal, const cv::Vec3d      rotationGlobal) {
	//this->Turn(rotationGlobal, 0);

	boost::lock_guard< boost::mutex > lock( mt_state_ ) ;
	const double droneYaw = AutoNavigator::toDroneYaw(rotationGlobal);
	droneYawTarget_ = droneYaw + 0;

	this->translationPrevious_ = translationGlobal;
	this->rotationPrevious_ = rotationGlobal;

	this->state_ = AutoNavigatorState::stateInactive;
}
bool AutoNavigator::getCommands(       DroneCommands& commands,
		const cv::Vec3d      rotationGlobal,
		const cv::Vec3d      translationGlobal ) {
	boost::lock_guard< boost::mutex > lock( mt_state_ ) ;
	if ( this->state_ == AutoNavigatorState::stateRotating
			){ //|| this->state_ == AutoNavigatorState::stateInactive) {
		//printing("AutoNavigator - rotating");
		// map angle difference to the interval [-180°,180°]
		const double droneYaw = AutoNavigator::toDroneYaw(rotationGlobal);
		double errorYaw = droneYawTarget_ - droneYaw ;
		errorYaw -= 360.0 * ( errorYaw >   180.0 ) ;
		errorYaw += 360.0 * ( errorYaw < - 180.0 ) ;

		// closed-loop control of yaw rotation
		if (fabs(errorYaw) < 10.0) {
			commands.hover();
			//this->state_ = AutoNavigatorState::stateInactive;
			this->state_ = AutoNavigatorState::stateInactive;
			//printing("back to state inactive");
			//TO DO: it need to set the command into DISENGAGE,
			//but right now I can't communicate between AutoNavigator and DroneAppBase
		}
		else {
			commands.controlAngZ( AutoNavigator::toRadian(errorYaw));
			// do nothing else except rotating
			commands.movement.linear.x = 0.0f ;
			commands.movement.linear.y = 0.0f ;
			commands.movement.linear.z = 0.0f ;
			commands.maneuver = DroneCommands::maneuverNone ;
		}

		const cv::Vec3d translDeltaGlobal = translationGlobal - this->translationPrevious_;
		std::cout<<translDeltaGlobal(0)<<" "
				<<translDeltaGlobal(1)<<" "
				<<translDeltaGlobal(2)<<" "<<std::endl;
//
//		printing(utilities::NumberToString(translDeltaGlobal(0))
//			+ utilities::NumberToString(translDeltaGlobal(1))
//			+ utilities::NumberToString(translDeltaGlobal(2)));
	}

	else if ( this->state_ == AutoNavigatorState::stateMoving
			|| this->state_ == AutoNavigatorState::stateInactive) {
//printing("state moving...");
		// closed-loop control of forward/backward and left/right motions based on visual odometry
		double errorForward = 0.0,
				errorLeft    = 0.0 ;

		//need to store previous translation.
		const cv::Vec3d translDeltaGlobal = translationGlobal - this->translationPrevious_;
		const cv::Vec3d translDeltaLocal = OdometryDrone::rotateTranslation( translDeltaGlobal, rotationGlobal( 1 ) ) ;
		errorForward = - translDeltaLocal( 2 ) ;
		errorLeft    =   translDeltaLocal( 0 ) ;
		commands.controlLinX( errorForward ) ;
		commands.controlLinY( errorLeft    ) ;

		// closed-loop control of altitude and yaw always based on on-board sensors
		const double altitudeBefore = - this->translationPrevious_(1),
				altitudeCurr   = -       translationGlobal(         1 ),
				yawBefore      = - this->rotationPrevious_(2),
				yawCurr        = -       rotationGlobal(            1 ) ;
		//this->altitudeDelta =
		double errorUp  = altitudeBefore - altitudeCurr + this->altitudeDelta,
				errorYaw = yawBefore      - yawCurr                            ;
		if( errorYaw >   CV_PI ) { errorYaw -= 2.0 * CV_PI ; }
		if( errorYaw < - CV_PI ) { errorYaw += 2.0 * CV_PI ; }
		commands.controlLinZ( errorUp  ) ;
		commands.controlAngZ( errorYaw ) ;

		// crank height control up if error is large
#if 1
		commands.controllerLinZ.disableIntegrator = false ;
		if( 0.0     < this->altitudeDelta
				&& errorUp > this->altitudeDelta * 0.25
				&& commands.movement.linear.z <   0.75f ) {
			commands.controllerLinZ.disableIntegrator = true ;
			commands.movement.linear.z =    0.75f ;
		}
		if( 0.0     > this->altitudeDelta
				&& errorUp < this->altitudeDelta * 0.25
				&& commands.movement.linear.z > - 0.75f ) {
			commands.controllerLinZ.disableIntegrator = true ;
			commands.movement.linear.z =  - 0.75f ;
		}
#endif
//
//		// show the errors
//		HAWAII_PRINT( errorForward ) ;
//		HAWAII_PRINT( errorLeft    ) ;
//		HAWAII_PRINT( errorUp      ) ;
//		HAWAII_PRINT( errorYaw     ) ;
//		std::cout << std::endl ; //*/

		// suppress any specific flight maneuver
		commands.maneuver = DroneCommands::maneuverNone ;

		// check if height change has been completed
		if( (fabs( errorForward ) < 0.20 * this->altitudeDelta ) // greater impact than others
				&& fabs( errorLeft    ) < 0.20 * this->altitudeDelta
				&& fabs( errorUp      ) < 0.20 * this->altitudeDelta
				&& fabs( errorYaw     ) < 5.0 / 180.0 * CV_PI ) {
			commands.hover();
			//this->state_ = AutoNavigatorState::stateInactive;
			printing("back to state inactive");
			// change state
			//this->timerHover.tic() ;
			//std::cout << "INFO: dense 3D: changed state to HOVER AFTER height change" << std::endl ;
		}
	}
	else {

		commands.hover();
	}

	//printf( "errorYaw = %5.2f, target = %5.2f, current = %5.2f\n", errorYaw, droneYawTarget_, AutoNavigator::toDroneYaw(rotationGlobal));

	//this->translationPrevious_ = translationGlobal;
//	this->rotationPrevious_ = rotationGlobal;

	//}
	//	if ( this->state_ == AutoNavigatorState::stateInactive) {
	//		//printing("AutoNavigator - state inactive")
	//		//commands.controlAngZ( 0 ) ;
	//		//set all the movement to zero and maneuverNone when there are nothing to do
	//		//I still don't know why: after I set the error to Zero and put it into the commands,
	//		//the Drone still turn in a slower speed!!! Except using commands.hover().
	//		commands.hover();
	//		printf( "current = %5.2f\n", AutoNavigator::toDroneYaw(rotationGlobal));
	//
	//	}
	return ( this->state_ != AutoNavigatorState::stateInactive ) ;
}

DroneProducer::DroneProducer(int n, int TimeBetweenImages):DroneAppBase(false)
//,mTimeBetweenImages(0, 0, 0, TimeBetweenImages)
{
	id_ = n;

	mTimeBetweenImages = boost::posix_time::milliseconds(TimeBetweenImages);
	mSendImageLastTime = boost::posix_time::microsec_clock::local_time();
	mSendImageDeltaTime = boost::posix_time::time_duration(0, 0, 0, 0);
}

DroneProducer::~DroneProducer() {
	// TODO Auto-generated destructor stub

}

boost::posix_time::ptime Glostart_time = boost::posix_time::microsec_clock::local_time();
void DroneProducer::processImageFront( const cv_bridge::CvImage imageFront ) {
	if (Gb_SystemState.get() != SystemState::activeState) {
		return;
	}

	boost::posix_time::ptime current_time = boost::posix_time::microsec_clock::local_time();
	boost::posix_time::time_duration delta_time = current_time - mSendImageLastTime;
	//std::cout << "DroneProducer::processImageFront:: boost::posix"<<std::endl;

	if (mSendImageDeltaTime + delta_time > mTimeBetweenImages) {
		mSendImageLastTime = current_time;
		cv::Mat visualization = imageFront.image ;

		MessageData msg;
		msg.mLapNo = this->id_;
		msg.mImg = visualization.clone();
		msg.mTimeStamp = boost::posix_time::microsec_clock::local_time();
		msg.mRotationVector = this->odoDrone.getRotation();
		msg.mTranslationVector = this->odoDrone.getTranslation();

		GloQueueData.push(msg);

		//TODO: do something with auto navigator
		//setup a navigator for both DroneProducer1 and 2
//		remoteNavigator_.getCommands(this->commands,
//				this->odoDrone.getRotation(),
//				this->odoDrone.getTranslation());
	}
	else {
		//printing("NO send image when delta time = " + utilities::NumberToString((mSendImageDeltaTime + delta_time).total_milliseconds()));
	}
}

void DroneProducer::processKeystrokes() {
	//	printing("DroneProducer::processKeystrokes()");
	//	if( this->keystates[ SDLK_ESCAPE ]) {
	//		GloConnect.setCloseConnection(true);
	//		MessageData msg = MessageData::createMessage(Command::CloseConnection);
	//		this->setData(msg);
	//		printing("stopping processKeystrokes");//: sent message close connection");
	//	}
}

//--------------------------------------------------------------------------

DroneProducer1::DroneProducer1(int n, int TimeBetweenImages):
				DroneProducer(n, TimeBetweenImages)

//				sparse3D2( this->focalLengthFrontU,
//						this->principalPointFrontU,
//						this->principalPointFrontV )
{
//	md_Stream1_ = NULL;
//	md_Stream2_ = NULL;
//
//	md_CurrentStream1_ = NULL;
//	md_CurrentStream2_ = NULL;

	this->consumerThread_ = boost::thread( &DroneProducer1::consume, this ) ;
}

DroneProducer1::~DroneProducer1() {
	if( this->consumerThread_.joinable() ) {
		this->consumerThread_.join() ;
	}
}

void DroneProducer1::processKeystrokes() {
	//stop connection when
	// Step 1: User press ESC button or remote press ESC button
	// Step 2: Set global variable Gb_SystemState to closedState
	// Step 3: Check state in sender and receiver according to each Drone
	//  For example: check in function session_server1::process_receiver and client 1 for Drone 1
	//               check in function session_server2::process_receiver and client 2 for Drone 2
	if( this->keystates[ SDLK_ESCAPE ]) {
		printing("change to stoped state!");
		Gb_SystemState.set(SystemState::closedState);
	}

	MessageData msg;
	Command cm = Command::NoCommand;

	if( this->keystates[ 		SDLK_1 		]) {cm = Command::Escape		; printing("Remote Escape		");}
	else if( this->keystates[ 	SDLK_2 		]) {cm = Command::TakeOff		; printing("Remote TakeOff		");}
	else if( this->keystates[ 	SDLK_3		]) {cm = Command::Landing		; printing("Remote Landing		");}
	else if( this->keystates[ 	SDLK_4 		]) {cm = Command::Engage		; printing("Remote Engage		");}
	else if( this->keystates[ 	SDLK_5 		]) {cm = Command::DisEngage		; printing("Remote DisEngage	");}

	else if( this->keystates[ 	SDLK_w 		]) {cm = Command::MoveForward	; printing("Remote MoveForward	");}
	else if( this->keystates[ 	SDLK_s 		]) {cm = Command::MoveBackward	; printing("Remote MoveBackward	");}
	else if( this->keystates[ 	SDLK_a 		]) {cm = Command::MoveLeft		; printing("Remote MoveLeft		");}
	else if( this->keystates[ 	SDLK_d 		]) {cm = Command::MoveRight		; printing("Remote MoveRight	");}
	else if( this->keystates[ 	SDLK_r 		]) {cm = Command::MoveUp		; printing("Remote MoveUp		");}
	else if( this->keystates[ 	SDLK_f 		]) {cm = Command::MoveDown		; printing("Remote MoveDown		");}

	else if( this->keystates[ 	SDLK_q 		]) {cm = Command::RotateLeft	; printing("Remote RotateLeft	");}
	else if( this->keystates[ 	SDLK_e 		]) {cm = Command::RotateRight	; printing("Remote RotateLeft	");}

	else if( this->keystates[ 	SDLK_b 		]) {
		cv::Vec3d rotationGlobal = this->odoDrone.getRotation();
		cv::Vec3d translationGlobal = this->odoDrone.getTranslation();
		remoteNavigator_.init(translationGlobal, rotationGlobal);
	}
	else if( this->keystates[ 	SDLK_0 		]) {
		printing("pressed number 0");
		//cm = Command::AutoRotate	;
		if (remoteNavigator_.checkState(AutoNavigatorState::stateInactive)) {
			cv::Vec3d rotationGlobal = this->odoDrone.getRotation();
			cv::Vec3d translationGlobal = this->odoDrone.getTranslation();
			remoteNavigator_.Turn(translationGlobal, rotationGlobal, 90);
			//remoteNavigator_.MoveUpDown(translationGlobal, rotationGlobal, 90);
			printing("turn	");
		}
	}
	else if( this->keystates[ 	SDLK_9 		]) {
			printing("pressed number 9");
			//cm = Command::AutoRotate	;
			if (remoteNavigator_.checkState(AutoNavigatorState::stateInactive)) {
				cv::Vec3d rotationGlobal = this->odoDrone.getRotation();
				cv::Vec3d translationGlobal = this->odoDrone.getTranslation();
				//remoteNavigator_.Turn(translationGlobal, rotationGlobal, 90);
				remoteNavigator_.MoveUpDown(translationGlobal, rotationGlobal, 0.5);
				printing("moving	");
			}
		}

	if (cm != Command::NoCommand) {
		msg = MessageData::createMessage(cm);
		GloQueueCommand.push(msg);
		//this->setCommand(cm);
	}
}


void DroneProducer1::consume() {
	printing("running DroneConsumer No " +utilities::NumberToString(id_));

	while (true) {
		if (Gb_SystemState.check(SystemState::closedState)) { break;}
		processor_.process();
	}

	printing("stopping DroneConsumer No " +utilities::NumberToString(id_));
}


//--------------------------------------------------------------------------

DroneProducer2::DroneProducer2(int n, int TimeBetweenImages):
				DroneProducer(n, TimeBetweenImages) {
	remoteNavigator_.setState(AutoNavigatorState::stateInactive);
}

DroneProducer2::~DroneProducer2() { }

void DroneProducer2::processKeystrokes() {
	//stop connection when
	// Step 1: User press ESC button or remote press ESC button
	// Step 2: Set global variable Gb_SystemState to closedState
	// Step 3: Check state in sender and receiver according to each Drone
	//  For example: check in function session_server1::process_receiver and client 1 for Drone 1
	//               check in function session_server2::process_receiver and client 2 for Drone 2
	if( this->keystates[ SDLK_ESCAPE ]) {
		printing("change to stoped state!");
		Gb_SystemState.set(SystemState::closedState);
	}

	MessageData msg;
	if (GloQueueCommand.try_pop(msg)) {
		//printing("get a command");

		boost::posix_time::ptime current_time = boost::posix_time::microsec_clock::local_time();
		boost::posix_time::time_duration diff = current_time - msg.mTimeStamp;
		//if (diff.total_milliseconds() > 2)
		{
			//drop packet
			printing("time for sending a command from PC 1: " + utilities::NumberToString(diff.total_milliseconds()));
			//printing("drop old packet!");
			//return;
		}

		this->setCommand(msg.mCommand);

		if( msg.isCommandMessage(Command::TakeOff)) {
			//trigger takeoff
			printing("Remote TakeOff");

		}
		else if( msg.isCommandMessage(Command::Landing)) {
			//trigger landing
			printing("Remote Landing");
		}
		else if( msg.isCommandMessage(Command::MoveForward)) {
			//trigger move forward
			printing("Remote MoveForward");
		}

		//engaged - mode
		else if( msg.isCommandMessage(Command::Engage)) {
			cv::Vec3d rotationGlobal = this->odoDrone.getRotation();
			cv::Vec3d translationGlobal = this->odoDrone.getTranslation();
			remoteNavigator_.init(translationGlobal, rotationGlobal);
		}
		else if( msg.isCommandMessage(Command::AutoRotate)) {
			printing("Remote AutoRotate");
			if (remoteNavigator_.checkState(AutoNavigatorState::stateInactive)) {
				cv::Vec3d rotationGlobal = this->odoDrone.getRotation();
				cv::Vec3d translationGlobal = this->odoDrone.getTranslation();
				remoteNavigator_.Turn(translationGlobal, rotationGlobal, 90);
				printing("rotating	");
			}
		}
	}

	remoteNavigator_.getCommands(this->commands,
			this->odoDrone.getRotation(),
			this->odoDrone.getTranslation());
}

//--------------------------------------------------------------------------
//DroneProducerServer
DroneProducerServer::DroneProducerServer(int n,
		const std::string& port,
		const std::string& hostLap2, const std::string& portLap2):
		network_base_infomation(n, port, hostLap2, portLap2) {
	printing(mportSrc + mhostDst + mportDst);
}

DroneProducerServer::~DroneProducerServer() {
	// TODO Auto-generated destructor stub
}

void DroneProducerServer::operator()() {
	try {
		boost::asio::io_service io_service;

		server s(io_service, utilities::StringToNumber(this->mportSrc), 1);
		client1 c(io_service, this->mhostDst, this->mportDst);

		boost::thread t_client1 = boost::thread(&client1::process_sender, boost::ref(c) ) ;
		t_client1.detach();

		while (!Gb_SystemState.check(SystemState::closedState)) {
			io_service.poll();
			//std::cout<<"DroneProducerServer check state"<<std::endl;
			//boost::this_thread::sleep(boost::posix_time::milliseconds(50));
		}

		s.socket_cancel();
		c.socket_cancel();

		std::cout<<"DroneProducerServer::operator() passed socketcancel"<<std::endl;
	} catch (std::exception& e)
	{
		printing("[ERROR] DroneProducerServer::operator (): " + std::string(e.what()));
	}
}

} /* namespace producer_thread */
