/*
 * processor.cpp
 *
 *  Created on: Nov 26, 2013
 *      Author: truongnt
 */

#include "processor.h"
#include "StitchingImage.h"
#include "hawaii/common/timer.h"

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/filter/zlib.hpp>
#include <fstream>

#include "base/cvmat_serialization.h"


namespace producer_consumer_thread {
ProcessorData2Drone::ProcessorData2Drone():
								sparse3D2( 500,
										300,
										300)
{
	sensor_msgs::CameraInfo cam_info;
	std::string camera_name;
	camera_calibration_parsers::readCalibration("ardrone_front1.yaml", camera_name, cam_info);
	std::cout <<"camera_info"
			<<cam_info <<std::endl;
	_process2Streams = new Process2Streams(cam_info.K[0], cam_info.K[2], cam_info.K[5]);
}

int CountListMessageData(std::list<MessageData> &listMD) {
	std::list<MessageData>::iterator iterator;
	int iCounter = 0;
	for (iterator = listMD.begin(); iterator != listMD.end(); ++iterator) {
		iCounter++;

	}
	return iCounter;
}

void RemoveOldMessageData(std::list<MessageData> &listMD, int iMaxTime) {
	std::list<MessageData>::iterator iterator_remover;
	boost::posix_time::time_duration diff1;

	iterator_remover = listMD.begin();
	boost::posix_time::ptime current_time = boost::posix_time::microsec_clock::local_time();
	while (iterator_remover != iterator_remover)
	{
		diff1 = current_time - iterator_remover->mTimeStamp;
		if (diff1.total_milliseconds() > iMaxTime) {
			listMD.erase(iterator_remover++);  // alternatively, i = items.erase(i);
		}
		else {
			break;
		}
	}
}

void ProcessorData2Drone::process() {
	MessageData msg;
	if (!GloQueueData.try_pop(msg)) {
		return;
	}
	int iMaxTime = 5000;

	int iLengh = CountListMessageData(listWaitingSynchronizeStream_);
	std::cout<<"before: "<<iLengh;
//	RemoveOldMessageData(listWaitingSynchronizeStream_, iMaxTime);
//	iLengh = CountListMessageData(listWaitingSynchronizeStream_);
//	std::cout<<" and after: "<<iLengh<<"remove old message"<<std::endl;

	{
		boost::posix_time::ptime current_time = boost::posix_time::microsec_clock::local_time();
		boost::posix_time::time_duration diff = current_time - msg.mTimeStamp;
		if (diff.total_milliseconds() > 1000) {
			//drop packet
			std::cout << "time for sending a packet from PC 2: " << utilities::NumberToString(diff.total_milliseconds())
				<< "drop!" <<std::endl;
			return;
		}
		if (msg.mLapNo == 2) {
			printing("time for sending a packet from PC 2: " + utilities::NumberToString(diff.total_milliseconds()));
		}
	}

	if (msg.isCommandMessage(Command::CloseConnection)) {
		cv::waitKey(5000);
		return;
	}

	//synchronize two streams into 1
	//condition: know that stream 1 is sent more frames/s than stream 2.
	if (md_Stream1_ == NULL || md_Stream2_ == NULL) {
		if (msg.mLapNo == 2) {
			md_Stream2_ = new MessageData(msg);
		}
		else {
			//Lapno = 1 --> add to waiting list
			listWaitingSynchronizeStream_.push_back(msg);
		}
	}

	//compare diff in order to find the best pair (the increasing of diff time)
	if (md_Stream1_ == NULL && md_Stream2_ != NULL) {
		std::list<MessageData>::iterator iterator;
		boost::posix_time::time_duration diff1;
		boost::posix_time::time_duration diff2;
		for (iterator = listWaitingSynchronizeStream_.begin(); iterator != listWaitingSynchronizeStream_.end(); ++iterator) {
			//update diff 1 if current iterator in the begining
			if (iterator == listWaitingSynchronizeStream_.begin()) {
				if (md_Stream2_->mTimeStamp > iterator->mTimeStamp) {
					diff1 = md_Stream2_->mTimeStamp - iterator->mTimeStamp;
				}
				else {
					diff1 = -(md_Stream2_->mTimeStamp - iterator->mTimeStamp);
				}
			}
			else {//update diff 2
				if (md_Stream2_->mTimeStamp > iterator->mTimeStamp) {
					diff2 = md_Stream2_->mTimeStamp - iterator->mTimeStamp;
				}
				else {
					diff2 = -(md_Stream2_->mTimeStamp - iterator->mTimeStamp);
				}
			}

			if (iterator != listWaitingSynchronizeStream_.begin() && diff1 < diff2) {
				--iterator;// get the first pair
				//std::cout<<"diff1"<< diff1<<std::endl;
				md_Stream1_ = new MessageData((*iterator));
				//remove used waiting list

				std::list<MessageData>::iterator iterator_remover;
				iterator_remover = listWaitingSynchronizeStream_.begin();
				while (iterator_remover != iterator)
				{
					listWaitingSynchronizeStream_.erase(iterator_remover++);  // alternatively, i = items.erase(i);
				}
				break;
			}

			if (iterator != listWaitingSynchronizeStream_.begin()) {
				diff1 = diff2;
			}
		}
	}

	if (md_Stream1_ != NULL && md_Stream2_ != NULL) {
		//add to current stream
		md_CurrentStream1_ = md_Stream1_;
		md_CurrentStream2_ = md_Stream2_;

		md_Stream1_ = NULL;
		md_Stream2_ = NULL;
		//merge into 1 and show image

		//TODO: show in 1 image
		this->process2ImageStreams(md_CurrentStream1_, md_CurrentStream2_);
	}
}

//if null, auto generate filePath depend on time
void Save2ImageStreams(MessageData * stream1, MessageData *stream2, std::string strFolderPath = "m") {
	boost::filesystem::path dir(strFolderPath);
	if (boost::filesystem::create_directory(dir)) {
		std::cout << "Success create folder " << strFolderPath << "\n";
	}

	strFolderPath += "/";
	//auto generate
	strFolderPath += utilities::NumberToString(utilities::getCurrentTimeDuration().total_milliseconds(), 12) + ".yml";

	std::cout<< strFolderPath << std::endl;
	//hawaii::tic();
//	{
//		FileStorage fs(strFolderPath, FileStorage::WRITE);
//
//		fs << "MessageData1" << *stream1;
//		fs << "MessageData2" << *stream2;
//
//		fs.release();
//		logging1("time to save first method", utilities::DoubleToString(hawaii::toc()));
//	}
//	{
//		std::cout<<stream1->mImg.cols<< " "<<stream1->mImg.rows<<std::endl;
//		std::cout<<stream2->mImg.cols<< " "<<stream2->mImg.rows<<std::endl;
//		cv::Mat resultImage = cv::Mat::zeros(stream1->mImg.rows * 2, stream1->mImg.cols + 3, stream1->mImg.type());
//		std::cout<<resultImage.cols<< " "<<resultImage.rows<<std::endl;
//		cv::Mat top(resultImage, cv::Rect(0, 0, stream1->mImg.cols, stream1->mImg.rows)),
//				bottom(resultImage, cv::Rect(0, stream1->mImg.rows, stream2->mImg.cols, stream2->mImg.rows));
//
//		std::cout<<resultImage.cols<< " "<<resultImage.rows<<std::endl;
//		stream1->mImg.copyTo(top);
//		stream2->mImg.copyTo(bottom);
//		HAWAII_IMSHOW( resultImage ) ;
//		cv::waitKey( 1 ) ;
//	}
	{//http://cheind.wordpress.com/2011/12/06/serialization-of-cvmat-objects-using-boost/
		namespace io = boost::iostreams;

		std::ofstream ofs(strFolderPath, std::ios::out | std::ios::binary);
		{ // use scope to ensure archive goes out of scope before stream
			io::filtering_streambuf<io::output> out;
			out.push(io::zlib_compressor(io::zlib::best_speed));
			out.push(ofs);

			boost::archive::binary_oarchive oa(out);
			oa << *stream1 << *stream2;
		}
		ofs.close();

//		logging1("time to save second method", utilities::DoubleToString(hawaii::toc()));
	}
//
//	std::cout << "rotation vector"<< stream1->mRotationVector <<std::endl
//				<< "translation vector" << stream1->mTranslationVector <<std::endl;
//
//	std::cout << "rotation vector"<< stream2->mRotationVector <<std::endl
//			<< "translation vector" << stream2->mTranslationVector <<std::endl;
//
//	std::cout<< "write done!"<<std::endl;
}
void ProcessorData2Drone::process2ImageStreams(
		MessageData *stream1,
		MessageData *stream2
		) {
	// keep original image for visualization, create GPU-suitable copies in color and (slightly blurred) grayscale
	//cv::Mat visualization = stream1->mImg;
//
//	//pre-processing for image stream 1
//	hawaii::AutoMat imageBGR1, imageGray1, imageGraySmooth1 ;
//	stream1->mImg.copyTo( imageBGR1.writeCPU( stream1->mImg.size(), stream1->mImg.type() ) ) ;
//	cv::cvtColor(      imageBGR1, imageGray1.writeCPU( imageBGR1.size(), CV_8UC1 ), cv::COLOR_BGR2GRAY ) ; // TODO multi-core
//
//	//pre-processing for image stream 2
//	hawaii::AutoMat imageBGR2, imageGray2, imageGraySmooth2 ;
//	stream2->mImg.copyTo( imageBGR2.writeCPU( stream2->mImg.size(), stream2->mImg.type() ) ) ;
//	cv::cvtColor(      imageBGR2, imageGray2.writeCPU( imageBGR2.size(), CV_8UC1 ), cv::COLOR_BGR2GRAY ) ; // TODO multi-core

	Save2ImageStreams(stream1, stream2, GlostrSaveFolder);

	_process2Streams->process2ImageStreams(stream1, stream2);

//	producer_consumer_thread::StichingImage stich;
//	stich.ApplyImage(stream1->mImg, stream2->mImg);
//
//	// go through sub-modules with decreasing priority
//	bool commandsSet = false ;
//
//	// continue visual odometry and sparse 3D reconstruction no matter if commands are already set, but do "corkscrew"
//	// flight with lowest priority: see note at dense 3D
//	if( 1 ) {
//		if( this->sparse3D2.processImageFront( imageGray1,//imageGraySmooth1,
//				stream1->mRotationVector, stream1->mTranslationVector
////				this->odoDrone.getRotation(),
////				this->odoDrone.getTranslation()
//				) ) {
//			//printing("first image for Sparse3D");
//		}
//		if( this->sparse3D2.processImageFront( imageGray2,//imageGraySmooth2,
//				stream2->mRotationVector, stream2->mTranslationVector
////				this->odoDrone.getRotation(),
////				this->odoDrone.getTranslation()
//				) ) {
//			//printing("second image for Sparse3D");
//		}
//
//		this->sparse3D2.visualize( visualization ) ;
//		if( !commandsSet ) {
//			//TODO: need implement something for command in here
////			commandsSet = this->sparse3D2.getCommands( this->commands,
//////					this->odoDrone.getTranslation()
////					translationStream1
////					) ;
//		}
//	}
//	//
//	//	// hover in place if no commands have been set for a given time
//	//	if( commandsSet ) { this->timerCommandsLastSet.tic() ; }
//	//	if( this->timerCommandsLastSet.toc() / 1000.0 > this->timeCommandsLastSetMax ) {
//	//		this->commands.reset() ;
//	//	}
//
//	// finally show the visualization image
//		HAWAII_IMSHOW( visualization ) ;
//	cv::waitKey( 1 ) ;
}

} /* namespace producer_thread */
