/*
 * sender_receiver.hpp
 *
 *  Created on: Aug 13, 2013
 *      Author: truongnt
 */

#ifndef SENDER_RECEIVER_HPP_
#define SENDER_RECEIVER_HPP_
//netstat --listen

#include <ctime>
#include <iostream>
#include <string>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/asio.hpp>

#include "base/base_services.h"

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>

#include "boost/thread/mutex.hpp"

//#include <boost/archive/text_oarchive.hpp>
//#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/filter/zlib.hpp>

using boost::asio::ip::tcp;
template <class Archive, class Stream, class Obj>
bool try_stream_next2(Archive &ar, const Stream &s, Obj &o)
{
  bool success = false;

  try {
    ar >> o;
    success = true;
  } catch (const boost::archive::archive_exception &e) {
    if (e.code != boost::archive::archive_exception::input_stream_error) {
      throw;
    }
  }

  return success;
}
enum ConnectionState {
	Step1,
	Step2,
	Step3,
	Step4,
	Step5,
	Step6,
	ConnectioEstablished,
	activeConnection,
	inactive
};

class session_server_base
{
public:
	session_server_base(boost::asio::io_service& io_service)
	: socket_(io_service)
	{
	}

	virtual ~session_server_base() {

	}

	tcp::socket& socket()
	{
		return socket_;
	}

	virtual void process_receiver(MessageData & msg) = 0;

	void asyn_getMessage()
	{
		try {
			//start to receive something
			printing("asyn_getMessage");
			socket_.async_read_some(boost::asio::buffer(&header_, sizeof(header_)),
					boost::bind(&session_server_base::handle_read_header, this,
							boost::asio::placeholders::error,
							boost::asio::placeholders::bytes_transferred));
		}
		catch (std::exception& e)
		{
			std::cerr << "asyn_getMessage " << e.what() << std::endl;
		}
	}

	void handle_read_header(const boost::system::error_code& error,
			size_t bytes_transferred)
	{
		try {
			if (!error)
			{
				printing("Body message: " + utilities::NumberToString(header_) + " bytes, transfered: " + utilities::NumberToString(bytes_transferred));
				if (sizeof(header_) == 4) // in case 32 bits, add more 4 bits to align with 64bits system
				{
					//printing("read more 4 bytes" );
					unsigned int a;
					boost::system::error_code error2;
					socket_.read_some(boost::asio::buffer( &a, sizeof(a) ), error2);
				}

				//--------------------------------------------------------------------------------------
				// read body
				remain_ = header_;

				socket_.async_read_some(streambuf_.prepare(remain_),
						boost::bind(&session_server_base::handle_read_body, this,
								boost::asio::placeholders::error,
								boost::asio::placeholders::bytes_transferred));
			}
			else
			{
				delete this;
			}
		}
		catch (std::exception& e)
		{
			std::cerr << "handle_read_header " << e.what() << std::endl;
		}
	}

	void handle_read_body(const boost::system::error_code& error,
			size_t bytes_transferred)
	{
		try {
			if (!error)
			{
				if (bytes_transferred > 0) {
					remain_ -= bytes_transferred;
					//http://www.boost.org/doc/libs/1_46_1/doc/html/boost_asio/reference/basic_streambuf.html
					//Move characters from the output sequence to the input sequence.
					streambuf_.commit(bytes_transferred);
					if (remain_ > 0) {
						socket_.async_read_some(streambuf_.prepare(remain_),
								boost::bind(&session_server_base::handle_read_body, this,
										boost::asio::placeholders::error,
										boost::asio::placeholders::bytes_transferred));
					}
					else {
						printing_with_color("Get a message successfully!", utilities::Color::green);
						// deserialize
//						std::istream is( &streambuf_ );
//						boost::archive::binary_iarchive ar( is );
////						boost::archive::text_iarchive ar( is );

						//http://cheind.wordpress.com/2011/12/06/serialization-of-cvmat-objects-using-boost/
						std::istream is( &streambuf_ );

						boost::iostreams::filtering_streambuf<boost::iostreams::input> in;
						in.push(boost::iostreams::zlib_decompressor());
						in.push(is);
						boost::archive::binary_iarchive ar( in );


						MessageData msg;
//						ar >> msg;
						try_stream_next2(ar, is, msg);
						this->process_receiver(msg);
					}
				}
			}
			else
			{
				delete this;
			}
		}
		catch (std::exception& e)
		{
			std::cerr << "handle_read_body " << e.what() << std::endl;
		}
	}

private:
	tcp::socket socket_;

	//for header
	size_t header_;

	//for body
	boost::asio::streambuf streambuf_;
	size_t remain_;
};

class session_server1: public session_server_base {
public:
	session_server1(boost::asio::io_service& io_service)
	: session_server_base(io_service)
	{
	}

	virtual void process_receiver(MessageData & msg) override {
		try {
			//stop connection when
			// Step 1: User press ESC button or remote press ESC button
			// Step 2: Set global variable Gb_SystemState to closedState
			// Step 3: Check state in sender and receiver according to each Drone
			//  For example: check in function session_server1::process_receiver and client 1 for Drone 1
			//               check in function session_server2::process_receiver and client 2 for Drone 2
			if (Gb_SystemState.check(SystemState::closedState)) { return; }

			//after finish asyn_getMessage, call this function for process that Message
			//printing("msg command: " + utilities::NumberToString(msg.mCommand));
			if (msg.isCommandMessage(Command::ClientSendInfoToServer)) {
				//			this->changeConnectionState(ConnectionState::activeConnection);
				//TODO: save this infomation
				this->asyn_getMessage();
				//			return;
			}
			else
				//		if (this->checkConnectionState(ConnectionState::activeConnection))
			{
				//			printing("activeConnection");
				GloQueueData.push(msg);//this line make things different
				this->asyn_getMessage();
				//			return;
			}
		}
		catch (std::exception& e)
		{
			std::cerr << "session_server1::process_receiver " << e.what() << std::endl;
		}
	}
};

class session_server2: public session_server_base {
public:
	session_server2(boost::asio::io_service& io_service)
	: session_server_base(io_service)
	{
	}
	virtual void process_receiver(MessageData & msg) override {
		try {
			//stop connection when
			// Step 1: User press ESC button or remote press ESC button
			// Step 2: Set global variable Gb_SystemState to closedState
			// Step 3: Check state in sender and receiver according to each Drone
			//  For example: check in function session_server1::process_receiver and client 1 for Drone 1
			//               check in function session_server2::process_receiver and client 2 for Drone 2
			if (Gb_SystemState.check(SystemState::closedState)) { return; }

			//after finish asyn_getMessage, call this function for process that Message
			//printing("msg command: " + utilities::NumberToString(msg.mCommand));
			if (msg.isCommandMessage(Command::ClientSendInfoToServer)) {
				//			this->changeConnectionState(ConnectionState::activeConnection);
				this->asyn_getMessage();
				//			return;
			}
			else
				//		if (this->checkConnectionState(ConnectionState::activeConnection))
			{
				//printing("activeConnection");
				//			printing("Message command: index: " + utilities::NumberToString(msg.mCommandIndex) + " command: "  + utilities::NumberToString(msg.mCommand));
				//GloQueueData.push(msg);
				GloQueueCommand.push(msg);//this line make things different
				this->asyn_getMessage();
				//			return;
				//
			}
		}
		catch (std::exception& e)
		{
			std::cerr << "session_server2::process_receiver " << e.what() << std::endl;
		}
	}
};

//create this class in DroneConsumer and DroneProducer
class server
{
protected:
	int id_;
	session_server_base* new_session_;
public:
	server(boost::asio::io_service& io_service, short port, int lapNo)
	: io_service_(io_service),
	  acceptor_(io_service, tcp::endpoint(tcp::v4(), port))
	{
		try {
			id_ = lapNo;

			printing("LISTENING port: " + utilities::NumberToString(port));

			this->create_session_server();

			acceptor_.async_accept(new_session_->socket(),
					boost::bind(&server::handle_accept, this, new_session_,
							boost::asio::placeholders::error));
		}
		catch (std::exception& e)
		{
			std::cerr << "server() " << e.what() << std::endl;
		}
	}

	void handle_accept(session_server_base* new_session,
			const boost::system::error_code& error)
	{
		try {
			if (!error)
			{
				printing("handle_accept successfully, there is 1 connection from client ");
				new_session->asyn_getMessage();
				//this->create_session_server();
//				acceptor_.async_accept(new_session_->socket(),
//						boost::bind(&server::handle_accept, this, new_session_,
//								boost::asio::placeholders::error));
			}
			else
			{
				//delete new_session;
				//show the error, continue listening util have the signal of stopping
//				this->create_session_server();
				acceptor_.async_accept(new_session_->socket(),
						boost::bind(&server::handle_accept, this, new_session_,
								boost::asio::placeholders::error));
			}
		}
		catch (std::exception& e)
		{
			std::cerr << "handle_accept " << e.what() << std::endl;
		}
	}
protected:
	void create_session_server() {
		if (id_ == 1) {
			new_session_ = new session_server1(io_service_);
		}
		else {
			new_session_ = new session_server2(io_service_);
		}
	}

protected:
	boost::asio::io_service& io_service_;
	tcp::acceptor acceptor_;

public:
	void socket_cancel() {
		this->acceptor_.cancel();
		std::cout<<"s.acceptor_.cancel();"<<std::endl;
		this->acceptor_.close();
		std::cout<<"s.acceptor_.close();"<<std::endl;
	}
};


//---------------------------------------------------------------------------------------------------------
//firstly, build up a sender in here to replace for the old one
class client
{
private:
	boost::asio::io_service& io_service_;
	tcp::socket socket_;
	std::string hostDst_, portDst_;

public:
	ProtectedVariable<bool> IsProcessingMessage_;
public:
	client(boost::asio::io_service& io_service,
			std::string hostDst, std::string portDst)
	: io_service_(io_service),
	  socket_(io_service)
	{
		try {
			connectionState_.set(ConnectionState::inactive);
			hostDst_ = hostDst;
			portDst_ = portDst;


			tcp::resolver resolver(io_service_);
			tcp::resolver::query query(hostDst_, portDst_);
			tcp::resolver::iterator iterator = resolver.resolve(query);

			tcp::endpoint endpoint = *iterator;
			socket_.async_connect(endpoint,
					boost::bind(&client::handle_connect, this,
							boost::asio::placeholders::error, ++iterator));

			IsProcessingMessage_.set(false);
		}
		catch (std::exception& e)
		{
			std::cerr << "client " << e.what() << std::endl;
		}
	}
	virtual ~client() {

	}

	ProtectedVariable<ConnectionState> connectionState_;
	//	mutable boost::mutex mutex_connectionState;
	//	bool checkConnectionState(const ConnectionState& currentState) {
	//		boost::mutex::scoped_lock lock(mutex_connectionState);
	//		if (this->mconnectionState == currentState) return true;
	//		return false;
	//	}
	//
	//	void changeConnectionState(const ConnectionState& currentState) {
	//		boost::mutex::scoped_lock lock(mutex_connectionState);
	//		this->mconnectionState = currentState;
	//	}


protected:
	void handle_connect(const boost::system::error_code& error,
			tcp::resolver::iterator endpoint_iterator)
	{
		try {
			if (!error)
			{
				printing("Connect to server successfully!");
				//::setSystemState(SystemState::activeState);
				Gb_SystemState.set(SystemState::activeState);
				connectionState_.set(ConnectionState::Step2);
				//this->process_sender();
				//IsProcessingMessage_.set(true);
			}
			else if (endpoint_iterator != tcp::resolver::iterator())
			{
				socket_.close();
				tcp::endpoint endpoint = *endpoint_iterator;
				socket_.async_connect(endpoint,
						boost::bind(&client::handle_connect, this,
								boost::asio::placeholders::error, ++endpoint_iterator));
			}
			else {
				printing("Can't connect! Re-connect");
				boost::this_thread::sleep(boost::posix_time::milliseconds(500));

				tcp::resolver resolver(io_service_);
				tcp::resolver::query query(hostDst_, portDst_);
				tcp::resolver::iterator iterator = resolver.resolve(query);

				tcp::endpoint endpoint = *iterator;
				socket_.async_connect(endpoint,
						boost::bind(&client::handle_connect, this,
								boost::asio::placeholders::error, ++iterator));
			}
		}
		catch (std::exception& e)
		{
			std::cerr << "handle_connect " << e.what() << std::endl;
		}
	}
public:
	virtual void process_sender() = 0;
protected:
	boost::asio::streambuf* streambuf_ptr_;
	std::ostream* os_ptr;
//	boost::archive::text_oarchive* ar_ptr;
	boost::archive::binary_oarchive* ar_ptr;
	size_t header_;
	void asyn_sendMessage(MessageData & msg)
	{
		try {
			//init variables.
			//		if (streambuf_ptr_) delete streambuf_ptr_;
			//		if (os_ptr) delete os_ptr;
			//		if (ar_ptr) delete ar_ptr;
//			streambuf_ptr_ = new boost::asio::streambuf();
//			os_ptr = new std::ostream(streambuf_ptr_);
////			ar_ptr = new boost::archive::text_oarchive(*os_ptr);
//			ar_ptr = new boost::archive::binary_oarchive(*os_ptr);
//			*ar_ptr << msg;

			//http://cheind.wordpress.com/2011/12/06/serialization-of-cvmat-objects-using-boost/
			streambuf_ptr_ = new boost::asio::streambuf();
			os_ptr = new std::ostream(streambuf_ptr_);
			{//http://stackoverflow.com/questions/1753469/how-to-hook-up-boost-serialization-iostreams-to-serialize-gzip-an-object-to
				//escape block to flush data into streambuf :)
				boost::iostreams::filtering_streambuf<boost::iostreams::output> fil_buf;
				fil_buf.push(boost::iostreams::zlib_compressor(boost::iostreams::zlib::best_speed));
				//fil_buf.push(*streambuf_ptr_);
				fil_buf.push(*os_ptr);

				ar_ptr = new boost::archive::binary_oarchive( fil_buf );

				*ar_ptr << msg;
			}

			size_t header_ = streambuf_ptr_->size();
			printing("body is " + utilities::NumberToString(header_) + " bytes" );

			// send header_ and buffer using scatter
			std::vector<boost::asio::const_buffer> buffers_;
			buffers_.push_back( boost::asio::buffer(&header_, sizeof(header_)) );
			//--------------------------------------------------------------------------------------
			if (sizeof(header_) == 4) // in case 32 bits, add more 4 bits to align with 64bits system
			{
				//printing("write more 4 bytes" );
				unsigned int a = 0;
				buffers_.push_back(boost::asio::buffer(&a, sizeof(a)));
			}
			//--------------------------------------------------------------------------------------
			buffers_.push_back( streambuf_ptr_->data() );






			boost::asio::async_write(socket_,
					buffers_,
					boost::bind(&client::handle_write, this,
							boost::asio::placeholders::error));
		}
		catch (std::exception& e)
		{
			std::cerr << "asyn_sendMessage " << e.what() << std::endl;
		}

	}

	void handle_write(const boost::system::error_code& error)
	{
		try {
			if (!error)
			{
				//printing("wrote " + utilities::NumberToString(rc_) + " bytes" );
				//streambuf_.consume(header_);
				printing_with_color("sent a message successfully!", utilities::Color::magenta);

				//process_sender();
				IsProcessingMessage_.set(false);
			}
			else
			{
				printing("[ERROR] handle_write: " + error.message());
				do_close();
			}
		}
		catch (std::exception& e)
		{
			std::cerr << "handle_write " << e.what() << std::endl;
		}
	}

	void do_close()
	{
		socket_.close();
	}

public:
	void socket_cancel() {
		socket_.cancel();
		socket_.close();
	}
};


class client1 :public client {
public:
	client1(boost::asio::io_service& io_service, std::string hostDst, std::string portDst)
	: client(io_service, hostDst, portDst)
	{
		CommandIndex_ = 0;
	}
public:
	virtual void process_sender() override;
private:
	int CommandIndex_;
};

class client2 :public client {
public:
	client2(boost::asio::io_service& io_service, std::string hostDst, std::string portDst)
	: client(io_service, hostDst, portDst)
	{

	}
public:
	virtual void process_sender() override;
};

#endif /* SENDER_RECEIVER_HPP_ */

